(define input-files
  '("all.scm"
    "utilities.scm"
    "random.scm"
    "sort.scm"
    "linear-algebra.scm"
    "sparse-linear-algebra.scm"
    "points.scm"
    "geometry.scm"
    "geometry-code.scm"
    "linear-elements.scm"
    "linear-elements-code.scm"
    "problem-descriptions.scm"
    "conjugate-gradient.scm"
    "intergrid.scm"
    "multigrid.scm"
    "parabolic.scm"
    "mmoc.scm"
    "obstacle.scm"
    "polygon.scm"
    "final-tests.scm"
    "final-tests-obstacle.scm"
    "test.scm"
    "test-parabolic.scm"
    "test-mmoc.scm"
    "tex.scm"
    "html-lib.scm"
    "describe.scm"))

(include "html-lib.scm")

(define (display-objects title os)
  (if (null? os)
      os
      (<li>
       title ":"
       (<ul>
	(map (lambda (o)
	       (let ((port (open-output-string)))
		 (define (my-display o)
		   (cond ((null? o) (display "()" port))
			 ((pair? o)
			  (display "(" port)
			  (my-display (car o))
			  (let loop ((rest (cdr o)))
			    (cond ((pair? rest)
				   (display " " port)
				   (my-display (car rest))
				   (loop (cdr rest)))
				  ((null? rest)
				   (display ")" port))
				  (else
				   (display " . " port)
				   (my-display rest)
				   (display ")" port)))))
			 (else
			  (display o port))))
		 (my-display o)
		 (<li> (<tt> (get-output-string port)))))
	     os)))))

(define (all-null? l)
  (or (null? l)
      (and (null? (car l))
	   (all-null? (cdr l)))))

(define (category-name c)
  (vector-ref c 0))

(define (category-recognizer c)
  (vector-ref c 1))

(define (category-extractor c)
  (vector-ref c 2))

(define categories
  ;; assumes that the form is a pair.
  ;; contains first the name, then the recognizer, then the extractor
  (list (vector "Classes"
		(lambda (form) (eq? (car form) 'define-class))
		(lambda (form) (cadr form)))     ; class name
	(vector "Generic functions"
		(lambda (form) (eq? (car form) 'define-generic))
		(lambda (form) (cadr form)))     ; name and argument list
	(vector "Methods"
		(lambda (form) (or (eq? (car form) 'define-method)
				   (eq? (car form) 'define-handy-method)))
		(lambda (form) (cadr form)))     ; name and argument list
	(vector "Functions"
		(lambda (form) (and (eq? (car form) 'define)
				    (pair? (cdr form))
				    (pair? (cadr form))))
		(lambda (form) (cadr form)))     ; name and argument list
	(vector "Other objects"
		(lambda (form) (and (eq? (car form) 'define)
				    (pair? (cdr form))
				    (symbol? (cadr form))))
		(lambda (form) (cadr form)))
	(vector "Pervasive macros"
		(lambda (form) (eq? (car form) 'define-meroon-macro))
		(lambda (form) (cadr form)))     ; name and argument list
	(vector "Local macros"
		(lambda (form) (or (eq? (car form) 'define-internal-meroon-macro)
				   (eq? (car form) 'define-macro)))
		(lambda (form) (cadr form)))
	(vector "Includes"
		(lambda (form) (eq? (car form) 'include))
		(lambda (form) (cadr form)))
	(vector "HTML tags"   ; name
		(lambda (form) (eq? (car form) 'define-tag))
		(lambda (form) (cadr form)))))   ; name


  (declare (not inline) (inlining-limit 0))
  (define (get-forms)
    (let loop ((files input-files) (results '()))
      (if (null? files)
	  (reverse results)
	  (let ((port (open-input-file (car files))))
	    (let inner ((forms '()) (form (read port)))
	      (if (eof-object? form)
		  (begin
		    (close-input-port port)
		    (loop (cdr files) (cons (cons (car files) (reverse forms)) results)))
		  (inner (cons form forms)
			 (read port))))))))
  (define (make-page all-forms)
    (<html>
     (<head> 
      (<title> 
       "Software for Numerical Methods for Partial Differential Equations"))
     (<body>
      (<center> 
       (<b> 
	(<font> size: "+4" 
		"Software for Numerical Methods for Partial Differential Equations")))
      (<p> 
       "
This software was developed for and by the students in "(<a> href: "http://www.math.purdue.edu/~lucier/615" "CS 615")", Numerical Methods for
Partial Differential Equations in the Spring semester of 2000.  Its goal is to implement the
finite element method in two dimensions.  It is written in Scheme, and uses some features of Gambit-C,
the Scheme->C compiler written by Marc Feeley.  It uses Meroon, developed by Christian Queinnec, as its
object system.  (The version of Meroon we used is available "
       (<a> href: "meroon.tgz" "here")
       ".)
The last two files are the files used to build this index page.")
      (<p>
       (<b> "Note: ")
       "
This software is a work in progress.  While much of it works as intended, much of it has not yet even
been tested.")
      (<p>
       "
The software is relatively fast.  For example, the sparse-matrix/vector multiply code
runs in 10 cycles per flop on our Compaq
DS20 clone (dual 500 MHz 21264s with 4 megabyte caches) after compiling the C code generated by\n"
       "Gambit-C with gcc-2.95.1 with the options \"-mcpu=ev6 -O2 -mieee -fPIC -fno-math-errno\".\n"
       "
The -fno-math-errno option, backported from the development compiler, causes the compiler to not set ERRNO
if sqrt returns a NaN.  The "
       (<a> href: "c-code" "same code written in C")
       "
, with the same
matrix structure, the same compiler, and the same compiler options, "
       (<b> "has exactly the same performance."))
      (<p> 
       "
Note added later in course: Jason Gower wrote code to number the vertices,
and hence the linear elements, along a space-filling curve.  The new
connectivity pattern is in connectivity2.gz.  With this connectivity pattern,
the scheme code runs in 8.25 cycles/flop, which is again just as fast as the
C version.")
      (<p>
       (<b> "Note: ")
       "
The performance of the code changes daily.
Final results will be posted at the end of the course.")
      (<p>
       "The code is coming along nicely and passes the tests in "
       (<tt> "final-tests.scm")
       ". 
Note that with the infrastructure that we have built, it is trivial to compute
things like the L2 and H1 norms of the error.")
      (<p>
       "
Timing and accuracy results are now available for the multigrid code applied
to problems of varying difficulty in the file "
       (<a> href: "mg-results"  (<b> "mg-results."))
       "
Some results on bigger problems can be found in "
       (<a> href: "mg-results2"  (<b> "mg-results2.")))
      (<p>
       "
Timing and accuracy results are now available for the conjugate-gradient code applied
to problems of varying difficulty in the file "
       (<a> href: "cg-results"  (<b> "cg-results.")))
      (<p>  "The file " (<a> href: "parabolic.scm" "parabolic.scm")
	   " contains a solver for parabolic differential equations,
and the file " (<a> href: "mmoc.scm" "mmoc.scm") " contains a solver for
transport-dominated diffusion problems.")
      (<p>  "A gzipped tarball of all the sources can be found "
(<a> href: "PUFS.tar.gz" "here")". A short article describing the process of software
development can be found "(<a> href: "http://www.math.purdue.edu/~lucier/615/abstract.pdf" "here")".")
      (<p> (<a> href: "meroon.tgz"  (<b> "meroon.tgz")))
      (<p> (<a> href: "Changelog"  (<b> "Changelog")))
      (map (lambda (file-name+forms)
	     (let ((file-name (car file-name+forms))
		   (forms (cdr file-name+forms)))
	       (let ((contents (map (lambda (c) '()) categories)))
		 (let loop ((forms forms))
		   (if (not (null? forms))
		       (let ((form (car forms)))
			 (if (pair? form)
			     (let inner ((categories categories)
					 (contents contents))
			       (if (null? categories)
				   (loop (cdr forms))
				   (let ((c (car categories)))
				     (if ((category-recognizer c) form)
					 (begin
					   (set-car! contents (cons ((category-extractor c) form)
								    (car contents)))
					   (loop (cdr forms)))
					 (inner (cdr categories) 
						(cdr contents))))))
			     (loop (cdr forms))))
		       (list (<p>
			(<a> href: file-name (<b> file-name)))
			(if (all-null? contents)
			    '()
			    (<ul>
			     (map (lambda (category contents)
				    (display-objects (category-name category)
						     (reverse contents)))
				  categories contents)))))))))
	   all-forms))))
  
  (define (display-page page)
    (with-output-to-file "index.html"
      (lambda ()
        (html-display page))))

(define (build-index)
  (let* ((all-forms (time (get-forms)))
         (page (time (make-page all-forms))))
    (time (display-page page))))
